/***************************************************************************

	adhoc.c

	PSP AdHoc functions.

***************************************************************************/

#include "emumain.h"
#include <pspsdk.h>
#include <psputilsforkernel.h>
#include <pspnet.h>
#include <pspnet_adhoc.h>
#include <pspnet_adhocctl.h>
#include <pspnet_adhocmatching.h>
#include <pspwlan.h>

#define NUM_ENTRIES			32

#define PSP_LISTING			1
#define PSP_SELECTED		2
#define PSP_SELECTING		3
#define PSP_WAIT_EST		4
#define PSP_ESTABLISHED		5

/***************************************************************************
	֐p[^ (̂ŁEEE)
***************************************************************************/

#define MATCHING_CREATE_PARAMS	\
	3,							\
	0xa,						\
	0x22b,						\
	0x800,						\
	0x2dc6c0,					\
	0x5b8d80,					\
	3,							\
	0x7a120,					\
	matchingCallback

#define MATCHING_START_PARAMS	\
	matchingId,					\
	0x10,						\
	0x2000,						\
	0x10,						\
	0x2000,						\
	strlen(matchingData) + 1,	\
	(char *)matchingData


/***************************************************************************
	[Jϐ
***************************************************************************/

static int Server;
static int pdpId;

static char g_mac[16];
static char g_mymac[16];
static int  g_unk1;
static int  g_matchEvent;
static int  g_matchOptLen;
static char g_matchOptData[1000];
static char g_matchingData[32];
static int  matchChanged;
static int  matchingId;

static struct psplist_t
{
	char name[48];
	char mac[6];
} psplist[NUM_ENTRIES];

static int max;
static int pos;


/***************************************************************************
	[J֐
***************************************************************************/

/*--------------------------------------------------------
	Xg
--------------------------------------------------------*/

static void ClearPspList(void)
{
	max = 0;
	pos = 0;
	memset(&psplist, 0, sizeof(psplist));
}


/*--------------------------------------------------------
	Xgɒǉ
--------------------------------------------------------*/

static int AddPsp(char *mac, char *name, int length)
{
	int i;

	if (max == NUM_ENTRIES) return 0;
	if (length == 1) return 0;

	for (i = 0; i < max; i++)
	{
		if (memcmp(psplist[i].mac, mac, 6) == 0)
			return 0;
	}

	memcpy(psplist[max].mac, mac, 6);

	if (length)
	{
		if (length < 47)
			strcpy(psplist[max].name, name);
		else
			strncpy(psplist[max].name, name, 47);
	}
	else
		psplist[max].name[0] = '\0';

	max++;

	return 1;
}


/*--------------------------------------------------------
	Xg폜
--------------------------------------------------------*/

static int DelPsp(char *mac)
{
	int i, j;

	for (i = 0; i < max; i++)
	{
		if (memcmp(psplist[i].mac, mac, 6) == 0)
		{
			if (i != max - 1)
			{
				for (j = i + 1; j < max; j++)
				{
					memcpy(psplist[j - 1].mac, psplist[j].mac, 6);
					strcpy(psplist[j - 1].name, psplist[j].name);
				}
			}

			if (pos == i) pos = 0;
			if (pos > i) pos--;
			max--;

			return 0;
		}
	}

	return -1;
}


/*--------------------------------------------------------
	Xg\
--------------------------------------------------------*/

static void DisplayPspList(int top, int rows)
{
	if (max == 0)
	{
		msg_printf("Waiting for another PSP to join\n");
	}
	else
	{
		int i;
		char temp[20];

		video_copy_rect(show_frame, draw_frame, &full_rect, &full_rect);

		draw_scrollbar(470, 26, 479, 270, rows, max, pos);

		for (i = 0; i < rows; i++)
		{
			if ((top + i) >= max) break;

			sceNetEtherNtostr((u8 *)psplist[top + i].mac, temp);

			if ((top + i) == pos)
			{
				uifont_print(24, 40 + (i + 2) * 17, UI_COLOR(UI_PAL_SELECT), temp);
				uifont_print(190, 40 + (i + 2) * 17, UI_COLOR(UI_PAL_SELECT), psplist[top + i].name);
			}
			else
			{
				uifont_print(24, 40 + (i + 2) * 17, UI_COLOR(UI_PAL_NORMAL), temp);
				uifont_print(190, 40 + (i + 2) * 17, UI_COLOR(UI_PAL_NORMAL), psplist[top + i].name);
			}
		}

		video_flip_screen(1);
	}
}


/*--------------------------------------------------------
	I𒆂PSP̏擾
--------------------------------------------------------*/

static int GetPspEntry(char *mac, char *name)
{
	if (max == 0) return -1;

	memcpy(mac, psplist[pos].mac, 6);
	strcpy(name, psplist[pos].name);

	return 1;
}


/*--------------------------------------------------------
	Matching callback
--------------------------------------------------------*/

static void matchingCallback(int unk1, int event, char *mac2, int optLen, char *optData)
{
	switch (event)
	{
	case MATCHING_JOINED:
		AddPsp(mac2, optData, optLen);
		break;

	case MATCHING_DISCONNECT:
		DelPsp(mac2);
		break;

	default:
		g_unk1        = unk1;
		g_matchEvent  = event;
		g_matchOptLen = optLen;
		strncpy(g_matchOptData, optData, optLen);
		memcpy(g_mac, mac2, sizeof(char) * 6);
		matchChanged = 1;
		break;
	}
}


/***************************************************************************
	AdHocC^tF[X֐
***************************************************************************/

/*--------------------------------------------------------
	W[̃[h
--------------------------------------------------------*/

int pspSdkLoadAdhocModules(void)
{
	int modID;

	modID = pspSdkLoadStartModule("flash0:/kd/ifhandle.prx", PSP_MEMORY_PARTITION_KERNEL);
	if (modID < 0)
		return modID;

	modID = pspSdkLoadStartModule("flash0:/kd/memab.prx", PSP_MEMORY_PARTITION_KERNEL);
	if (modID < 0)
		return modID;

	modID = pspSdkLoadStartModule("flash0:/kd/pspnet_adhoc_auth.prx", PSP_MEMORY_PARTITION_KERNEL);
	if (modID < 0)
		return modID;

	modID = pspSdkLoadStartModule("flash0:/kd/pspnet.prx", PSP_MEMORY_PARTITION_USER);
	if (modID < 0)
		return modID;
	else
		pspSdkFixupImports(modID);

	modID = pspSdkLoadStartModule("flash0:/kd/pspnet_adhoc.prx", PSP_MEMORY_PARTITION_USER);
	if (modID < 0)
		return modID;
	else
		pspSdkFixupImports(modID);

	modID = pspSdkLoadStartModule("flash0:/kd/pspnet_adhocctl.prx", PSP_MEMORY_PARTITION_USER);
	if (modID < 0)
		return modID;
	else
		pspSdkFixupImports(modID);

	modID = pspSdkLoadStartModule("flash0:/kd/pspnet_adhoc_matching.prx", PSP_MEMORY_PARTITION_USER);
	if (modID < 0)
		return modID;
	else
		pspSdkFixupImports(modID);

	sceKernelDcacheWritebackAll();
	sceKernelIcacheInvalidateAll();

	return 0;
}


/*--------------------------------------------------------
	vOXo[
--------------------------------------------------------*/

static int progress_total;
static int progress_current;

static void init_progress(int progress, int total, const char *text)
{
	char buf[32];

	load_background();

	progress_current = progress;
	progress_total   = total;

	video_copy_rect(work_frame, draw_frame, &full_rect, &full_rect);

	small_icon(6, 3, UI_COLOR(UI_PAL_TITLE), ICON_SYSTEM);

	sprintf(buf, "AdHoc - %s", game_name);
	uifont_print(32, 5, UI_COLOR(UI_PAL_TITLE), buf);

	draw_dialog(240-158, 136-26, 240+158, 136+26);

	if (text) uifont_print_center(118, 255,255,255, text);

	boxfill(240-151, 138+2, 240+151, 138+14, 0, 0, 0);

	video_copy_rect(draw_frame, work_frame, &full_rect, &full_rect);

	draw_battery_status(1);

	if (progress)
	{
		int width = (progress * 100 / progress_total) * 3;

		boxfill(240-150, 138+3, 240-150+width-1, 138+13, 128, 128, 128);
	}

	video_flip_screen(1);
}

static void update_progress(void)
{
	int width = (++progress_current * 100 / progress_total) * 3;

	show_background();

	draw_battery_status(1);

	boxfill(240-150, 138+3, 240-150+width-1, 138+13, 128, 128, 128);

	video_flip_screen(1);
}


/*--------------------------------------------------------
	
--------------------------------------------------------*/

int adhocInit(const char *matchingData)
{
	struct productStruct product;
	int error = 0, state = 0;
	char mac[20], buf[256];

	Server = 0;

	g_unk1        = 0;
	g_matchEvent  = 0;
	g_matchOptLen = 0;
	matchChanged  = 0;
	memset(g_mac, 0, sizeof(g_mac));
	memset(g_mymac, 0, sizeof(g_mymac));

	sprintf(product.product, "CPS200%d%d%d", VERSION_MAJOR, VERSION_MINOR, VERSION_BUILD);
	product.unknown = 0;

	ClearPspList();

	if (strlen(matchingData) == 0)
		return -1;

	strcpy(g_matchingData, matchingData);

	init_progress(0, 10, "Connecting...");

	if ((error = sceNetInit(0x20000, 0x20, 0x1000, 0x20, 0x1000)) == 0)
	{
		update_progress();
		if ((error = sceNetAdhocInit()) == 0)
		{
			update_progress();
			if ((error = sceNetAdhocctlInit(0x2000, 0x20, &product)) == 0)
			{
				update_progress();
				if ((error = sceNetAdhocctlConnect((int *)"")) == 0)
				{
					update_progress();
					do
					{
						if ((error = sceNetAdhocctlGetState(&state)) != 0) break;
						sceKernelDelayThread(1000000/60);

					} while (state != 1);

					if (!error)
					{
						update_progress();

						sceWlanGetEtherAddr(mac);
						update_progress();

						if ((pdpId = sceNetAdhocPdpCreate(mac, 0x309, 0x400, 0)) > 0)
						{
							update_progress();
							if ((error = sceNetAdhocMatchingInit(0x20000)) == 0)
							{
								update_progress();
								if ((matchingId = sceNetAdhocMatchingCreate(MATCHING_CREATE_PARAMS)) >= 0)
								{
									update_progress();
									if ((error = sceNetAdhocMatchingStart(MATCHING_START_PARAMS)) == 0)
									{
										init_progress(10, 10, "Connected.");
										sceKernelDelayThread(1000000);
										return 0;
									}
									sceNetAdhocMatchingDelete(matchingId);
								}
								error = 2;
								sceNetAdhocMatchingTerm();
							}
							sceNetAdhocPdpDelete(pdpId, 0);
						}
						error = 1;
					}
					sceNetAdhocctlDisconnect();
				}
				sceNetAdhocctlTerm();
			}
			sceNetAdhocTerm();
		}
		sceNetTerm();
	}

	switch (error)
	{
	case 1:  sprintf(buf, "failed. (PDP ID = %08x)", pdpId); break;
	case 2:  sprintf(buf, "failed. (Matching ID = %08x)", matchingId); break;
	default: sprintf(buf, "failed. (Error Code = %08x)", error); break;
	}

	init_progress(progress_current, 10, buf);

	pad_wait_clear();
	pad_wait_press(PAD_WAIT_INFINITY);

	return -1;
}


/*--------------------------------------------------------
	ؒf
--------------------------------------------------------*/

int adhocTerm(void)
{
	init_progress(0, 5, "Disconnecting...");

	sceNetAdhocctlDisconnect();
	update_progress();

	sceNetAdhocPdpDelete(pdpId, 0);
	update_progress();

	sceNetAdhocctlTerm();
	update_progress();

	sceNetAdhocTerm();
	update_progress();

	sceNetTerm();

	init_progress(5, 5, "Disconnected.");
	sceKernelDelayThread(1000000);

	return 0;
}

/*--------------------------------------------------------
	ؒf
--------------------------------------------------------*/

static void adhocDisconnect(void)
{
	init_progress(0, 8, "Disconnecting...");

	sceNetAdhocMatchingStop(matchingId);
	update_progress();

	sceNetAdhocMatchingDelete(matchingId);
	update_progress();

	sceNetAdhocMatchingTerm();
	update_progress();

	sceNetAdhocctlDisconnect();
	update_progress();

	sceNetAdhocPdpDelete(pdpId, 0);
	update_progress();

	sceNetAdhocctlTerm();
	update_progress();

	sceNetAdhocTerm();
	update_progress();

	sceNetTerm();

	init_progress(8, 8, "Disconnected.");
	sceKernelDelayThread(1000000);
}

/*--------------------------------------------------------
	SSIDw肵čĐڑ
--------------------------------------------------------*/

int adhocReconnect(char *ssid)
{
	int error = 0, state = 1;
	char mac[20], buf[256];

	init_progress(0, 6, "Disconnecting...");

	sceNetAdhocMatchingStop(matchingId);
	update_progress();

	sceNetAdhocMatchingDelete(matchingId);
	update_progress();

	sceNetAdhocMatchingTerm();
	update_progress();

	sceNetAdhocPdpDelete(pdpId, 0);
	update_progress();

	sceNetAdhocctlDisconnect();
	update_progress();

	do
	{
		if ((error = sceNetAdhocctlGetState(&state)) != 0) break;
			sceKernelDelayThread(1000000/60);
	} while (state == 1);

	init_progress(6, 6, "Disconnected.");
	sceKernelDelayThread(1000000);

	init_progress(0, 4, "Connecting...");

	if ((error = sceNetAdhocctlConnect((int *)ssid)) == 0)
	{
		update_progress();
		do
		{
			if ((error = sceNetAdhocctlGetState(&state)) != 0) break;
			sceKernelDelayThread(1000000/60);
		} while (state != 1);

		if (!error)
		{
			update_progress();

			sceWlanGetEtherAddr(mac);
			memcpy(g_mymac, mac, 6);
			update_progress();

			if ((pdpId = sceNetAdhocPdpCreate(mac, 0x309, 0x800, 0)) > 0)
			{
				init_progress(4, 4, "Connected.");
				sceKernelDelayThread(1000000);
				return 0;
			}
			error = 1;
		}
		sceNetAdhocctlDisconnect();

		if (state == 1)
		{
			do
			{
				if ((error = sceNetAdhocctlGetState(&state)) != 0) break;
				sceKernelDelayThread(1000000/60);
			} while (state == 1);
		}
	}

	sceNetAdhocctlTerm();
	sceNetAdhocTerm();
	sceNetTerm();

	switch (error)
	{
	case 1:  sprintf(buf, "failed. (PDP ID = %08x)", pdpId); break;
	default: sprintf(buf, "failed. (Error Code = %08x)", error); break;
	}

	init_progress(progress_current, 4, buf);

	pad_wait_clear();
	pad_wait_press(PAD_WAIT_INFINITY);

	return -1;
}


/*--------------------------------------------------------
	ڑ̑I
--------------------------------------------------------*/

int adhocSelect(void)
{
	int top = 0;
	int rows = 11;
	int currentState = PSP_LISTING;
	int prev_max = 0;
	int update = 1;
	char mac[16];
	char name[64];
	char temp[64];
	char ssid[10];
	char title[32];

	sprintf(title, "AdHoc - %s", game_name);
	msg_screen_init(title);

	while (1)
	{
		pad_update();

		msg_set_text_color(0xffff);

		switch (currentState)
		{
		case PSP_LISTING:
			Server = 0;
			if (update)
			{
				msg_screen_init(title);
				msg_printf("Select a server to connect to, or " FONT_TRIANGLE " to return.\n");
				msg_printf("\n");
				DisplayPspList(top, rows);
				update = 0;
			}
			if (pad_pressed(PSP_CTRL_UP))
			{
				if (pos > 0) pos--;
				update = 1;
			}
			else if (pad_pressed(PSP_CTRL_DOWN))
			{
				if (pos < max - 1) pos++;
				update = 1;
			}
			else if (pad_pressed(PSP_CTRL_CIRCLE))
			{
				if (GetPspEntry(mac, name) > 0)
				{
					if (strcmp(name, g_matchingData) == 0)
					{
						currentState = PSP_SELECTING;
						sceNetAdhocMatchingSelectTarget(matchingId, mac, 0, 0);
						update = 1;
					}
				}
			}
			else if (pad_pressed(PSP_CTRL_TRIANGLE))
			{
				msg_set_text_color(0xffffffff);
				adhocDisconnect();
				pad_wait_clear();
				return -1;
			}
			if (matchChanged)
			{
				if (g_matchEvent == MATCHING_SELECTED)
				{
					memcpy(mac, g_mac, 6);
					strcpy(name, g_matchOptData);
					currentState = PSP_SELECTED;
				}
				update = 1;
			}
			break;

		case PSP_SELECTING:
			if (update)
			{
				msg_screen_init(title);
				sceNetEtherNtostr((u8 *)mac, temp);
				msg_printf("Waiting for %s to accept the connection.\n", temp);
				msg_printf("To cancel press " FONT_CROSS ".\n");
				update = 0;
			}
			if (pad_pressed(PSP_CTRL_CROSS))
			{
				sceNetAdhocMatchingCancelTarget(matchingId, mac);
				currentState = PSP_LISTING;
				update = 1;
			}
			if (matchChanged)
			{
				switch (g_matchEvent)
				{
				case MATCHING_SELECTED:
					sceNetAdhocMatchingCancelTarget(matchingId, mac);
					break;

				case MATCHING_ESTABLISHED:
					currentState = PSP_ESTABLISHED;
					break;

				case MATCHING_REJECTED:
					currentState = PSP_LISTING;
					break;
				}
				update = 1;
			}
			break;

		case PSP_SELECTED:
			Server = 1;
			if (update)
			{
				msg_screen_init(title);
				sceNetEtherNtostr((u8 *)mac, temp);
				msg_printf("%s has requested a connection.\n", temp);
				msg_printf("To accept the connection press " FONT_CIRCLE ", to cancel press " FONT_CROSS ".\n");
				update = 0;
			}
			if (pad_pressed(PSP_CTRL_CROSS))
			{
				sceNetAdhocMatchingCancelTarget(matchingId, mac);
				currentState = PSP_LISTING;
				update = 1;
			}
			else if (pad_pressed(PSP_CTRL_CIRCLE))
			{
				sceNetAdhocMatchingSelectTarget(matchingId, mac, 0, 0);
				currentState = PSP_WAIT_EST;
				update = 1;
			}
			if (matchChanged)
			{
				if (g_matchEvent == MATCHING_CANCELED)
				{
					currentState = PSP_LISTING;
				}
				update = 1;
			}
			break;

		case PSP_WAIT_EST:
			if (matchChanged)
			{
				if (g_matchEvent == MATCHING_ESTABLISHED)
				{
					currentState = PSP_ESTABLISHED;
				}
				update = 1;
			}
			break;
		}

		matchChanged = 0;
		if (currentState == PSP_ESTABLISHED)
			break;

		if (top > max - rows) top = max - rows;
		if (top < 0) top = 0;
		if (pos >= top + rows) top = pos - rows + 1;
		if (pos < top) top = pos;

		if (max != prev_max)
		{
			prev_max = max;
			update = 1;
		}

		sceDisplayWaitVblankStart();
	}

	msg_set_text_color(0xffffffff);

	if (Server) sceWlanGetEtherAddr(mac);

	sceNetEtherNtostr((u8 *)mac, temp);

	ssid[0] = temp[ 9];
	ssid[1] = temp[10];
	ssid[2] = temp[12];
	ssid[3] = temp[13];
	ssid[4] = temp[15];
	ssid[5] = temp[16];
	ssid[6] = '\0';

	if (adhocReconnect(ssid) < 0)
		return -1;

	return (Server ? 1 : 0);
}


/*--------------------------------------------------------
	f[^M (჌x)
--------------------------------------------------------*/

int adhocSend(void *buffer, int length)
{
	int error_code = sceNetAdhocPdpSend(pdpId, g_mac, 0x309, buffer, length, 0, 1);

	return ((error_code < 0) ? error_code : length);
}


/*--------------------------------------------------------
	f[^M (჌x)
--------------------------------------------------------*/

int adhocRecv(void *buffer, int length)
{
	int error_code = 0;
	int pdpStatLength = sizeof(pdpStatStruct);
	int port;
	pdpStatStruct pdpStat;

	if (sceNetAdhocGetPdpStat(&pdpStatLength, &pdpStat) >= 0)
	{
		if (pdpStat.rcvdData == (u32)length)
		{
			if (memcmp(pdpStat.mac, g_mymac, 6) == 0)
			{
				port = pdpStat.port;
				if ((error_code = sceNetAdhocPdpRecv(pdpId, pdpStat.mac, &port, buffer, &length, 0, 1)) >= 0)
				{
					return pdpStat.rcvdData;
				}
			}
		}
	}
	return error_code;
}


/*--------------------------------------------------------
	f[^M҂
--------------------------------------------------------*/

int adhocRecvBlocked(void *buffer, int length)
{
	TICKER limit;
	int error_code = 0;

	limit = ticker() + 1000000*30;

	while ((error_code = adhocRecv(buffer, length)) == 0)
	{
		sceKernelDelayThread(200);
		if (ticker() >= limit) return -1;
	}

	return error_code;
}


/*--------------------------------------------------------
	f[^𑗐MAackM܂ő҂
--------------------------------------------------------*/

int adhocSendRecvAck(void *buffer, int length)
{
	int error_code = 0;
	int ack_data   = 0;
	int tempLen    = length;
	int sentCount  = 0;
	u8 *buf = (u8 *)buffer;

	do
	{
		if (tempLen > 0x400) tempLen = 0x400;

		if ((error_code = adhocSend(buf, tempLen)) < 0)
			return error_code;

		if ((error_code = adhocRecvBlocked(&ack_data, sizeof(int))) != tempLen)
			return error_code;

		buf += 0x400;
		sentCount += 0x400;
		tempLen = length - sentCount;
	} while (sentCount < length);

	return error_code;
}


/*--------------------------------------------------------
	ackMĂf[^𑗐M
--------------------------------------------------------*/

int adhocRecvSendAck(void *buffer, int length)
{
	int error_code = 0;
	int ack_data   = 1;
	int tempLen    = length;
	int rcvdCount  = 0;
	u8 *buf = (u8 *)buffer;

	do
	{
		if (tempLen > 0x400) tempLen = 0x400;

		if ((error_code = adhocRecvBlocked(buf, tempLen)) != tempLen)
			return error_code;

		if ((error_code = adhocSend(&ack_data, sizeof(int))) < 0)
			return error_code;

		buf += 0x400;
		rcvdCount += 0x400;
		tempLen = length - rcvdCount;
	} while (rcvdCount < length);

	return error_code;
}
